/*
   Copyright 2016 Toby Kenney

   This file is part of SimplePlot.

   SimplePlot is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   SimplePlot is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SimplePlot.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.io.*;
import java.lang.Math.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.Graphics.*;
import java.awt.Graphics2D.*;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseMotionAdapter;
import java.util.Scanner;

class SquareMatrix{
    int sz;
    double[][] data;

    public SquareMatrix(double[][] dt,int dim){
	sz=dim;
	data=dt;
    }

    public double[] solve(double[] d){
	/*	for(int i=0;i<sz;i++){
	    for(int j=0;j<sz;j++){
		System.out.print(data[i][j]+" ");
	    }
	    System.out.println("  "+d[i]);
	}
	System.out.println("");*/
	double[][] temp=new double[sz][sz];
	double[]tempd=new double[sz];
	for(int i=0;i<sz;i++){
	    tempd[i]=d[i];
	    for(int j=0;j<sz;j++){
		temp[i][j]=data[i][j];
	    }
	}
	for(int i=0;i<sz;i++){
	    int pivot=i;
	    double max=Math.abs(temp[i][i]);
	    for(int j=i+1;j<sz;j++){
		if(Math.abs(temp[i][j])>max){
		    max=Math.abs(temp[i][j]);
		    pivot=j;
		}
	    }
	    if(pivot!=i){
		//swap rows
		for(int j=i;j<sz;j++){
		    double tmpcopy=temp[i][j];
		    temp[i][j]=temp[pivot][j];
		    temp[pivot][j]=tmpcopy;
		}
		double tmpcopy=tempd[i];
		tempd[i]=tempd[pivot];
		tempd[pivot]=tmpcopy;
	    }
	    for(int j=i+1;j<sz;j++){
		double factor=temp[j][i]/temp[i][i];
		temp[j][i]=0;
		for(int k=i+1;k<sz;k++){
		    temp[j][k]-=temp[i][k]*factor;
		};
		tempd[j]-=tempd[i]*factor;
	    }
	}

	/*	for(int i=0;i<sz;i++){
	    for(int j=0;j<sz;j++){
		System.out.print(temp[i][j]+" ");
	    }
	    System.out.println("  "+tempd[i]);
	}
	System.out.println("");*/

	double[] ans=new double[sz];
	for(int i=sz-1;i>=0;i--){
	    double sum=tempd[i];
	    for(int j=i+1;j<sz;j++){
		sum-=temp[i][j]*ans[j];
	    }
	    ans[i]=sum/temp[i][i];
	}
	return ans;
    }
}

class metadata{
    int ndata;
    int nummeta;
    String[] names;
    String[][] metaclasses;
    int[][] classes;

    int numlinks;
    int[] links;
    Color[] linkcolours;
    int[] linktype;


    metadata(File f,int nd){
	ndata=nd;
	Scanner stream=null;
	try{
	    stream=new Scanner(new BufferedReader(new FileReader(f)));
	} catch(FileNotFoundException e) {
	    System.err.println("File: "+f.getName()+" not found!");
	} 
	try{
	    nummeta=stream.nextInt();
	    String[][] tempcl=new String[nummeta][ndata];
	    int[][] tempclno=new int[nummeta][ndata];
	    int[] templen=new int[nummeta];
	    names=new String[ndata];
	    for(int i=0;i<nummeta;i++){
		templen[i]=0;
	    }
	    for(int i=0;i<ndata;i++){
		names[i]=stream.next();
		for(int j=0;j<nummeta;j++){
		    String mt=stream.next();
		    tempclno[j][i]=-1;
		    for(int k=0;k<templen[j];k++){
			if(mt.equals(tempcl[j][k])){
			    tempclno[j][i]=k;
			}
		    }
		    if(tempclno[j][i]==-1){
			tempclno[j][i]=templen[j];
			tempcl[j][templen[j]]=mt;
			templen[j]++;
		    }
		}
	    }
	    classes=tempclno;
	    metaclasses=tempcl;
	    //Test whether the file has link data
	    if(stream.hasNext()){
		String mt=stream.next();
		System.out.println(mt);

		if(mt.equals("LINKS")){
		    //file includes link data
		    //first number should be number of links		   
		    //link data is in the format start no end no linktype.
		    
		    //		    System.out.println("Found link data");
		   

		    numlinks=stream.nextInt();
		    

		    links=new int[numlinks*2];
		    linktype=new int[numlinks];
		    String[] lt=new String[numlinks];
		    for(int i=0;i<numlinks;i++){
			//read a link from the file
			String ln1=stream.next();
			String ln2=stream.next();
			links[2*i]=-1;
			links[2*i+1]=-1;
			for(int j=0;j<ndata;j++){
			    if(ln1.equals(names[j])){
				links[2*i]=j;
			    }
			    if(ln2.equals(names[j])){
				links[2*i+1]=j;
			    }
			}
			if(links[2*i]==-1){
			    //no match
			    links[2*i]=Integer.parseInt(ln1)-1;
			}
			if(links[2*i+1]==-1){
			    //no match
			    links[2*i+1]=Integer.parseInt(ln2)-1;
			}
			lt[i]=stream.next();
			//			System.out.println(names[links[2*i]]+" "+names[links[2*i+1]]+" "+lt[i]);
		    }
		    //		    linkcolours;
		    //		    linktype;
		    int numlinktypes=0;
		    String[] ltlist=new String[numlinks];
		    for(int i=0;i<numlinks;i++){
			int ltno=-1;
			for(int j=0;j<numlinktypes;j++){
			    if(lt[i].equals(ltlist[j])){
				ltno=j;
				break;
			    }
			}
			if(ltno==-1){
			    //new link type
			    ltlist[numlinktypes]=lt[i];
			    linktype[i]=numlinktypes;
			    numlinktypes++;			    
			}else{
			    linktype[i]=ltno;
			}
		    }
		    //Make list of link colours
		    linkcolours=new Color[numlinktypes];
		    int pred=80;
		    int pblue=80;
		    int pgreen=80;
		    for(int i=0;i<numlinktypes;i++){
			linkcolours[i]=new Color((pred*(i))%256,(pgreen*(i))%256,(pblue*(i))%256);
		    }
		}
	    }
	    //	    int tmpdim=stream.next();
	    stream.close();
	}finally{
	}	
    }
}

class Splot extends JPanel{
    int dim;
    int ndata;
    double[][] coeffs;
    double[][] points;
    double ptsize;
    int[] clss;
    Color[] colours;
    int moving;
    int width;
    int height;
    SimplePlot parent;
    metadata meta;
    int metanum;
    int numgreenpoints;

    public Splot(int n,int nd,SimplePlot p){
	parent=p;
	setBorder(BorderFactory.createLineBorder(Color.black));
	dim=n;
	ndata=nd;
	width=500;
	height=350;
	moving=-1;
	numgreenpoints=0;

	addMouseListener(new MouseAdapter() {
		public void mousePressed(MouseEvent e) {
		    mpress(e.getX(),e.getY());
		}
		public void mouseReleased(MouseEvent e) {
		    moving=-1;
		}
	    });

        addMouseMotionListener(new MouseAdapter() {
		public void mouseDragged(MouseEvent e) {
		    mdrag(e.getX(),e.getY());
		}
		public void mouseMoved(MouseEvent e) {
		    mmove(e.getX(),e.getY());
		}
	    });
    }

    public Splot(File f,SimplePlot p){
	parent=p;
	setBorder(BorderFactory.createLineBorder(Color.black));
	width=500;
	height=350;
	ptsize=5;
	moving=-1;
	this.loadData(f);

	addMouseListener(new MouseAdapter() {
		public void mousePressed(MouseEvent e) {
		    mpress(e.getX(),e.getY());
		}
		public void mouseReleased(MouseEvent e) {
		    moving=-1;
		}
	    });

        addMouseMotionListener(new MouseAdapter() {
		public void mouseDragged(MouseEvent e) {
		    mdrag(e.getX(),e.getY());
		}
		public void mouseMoved(MouseEvent e) {
		    mmove(e.getX(),e.getY());
		}
	    });
    }

    public Splot(SimplePlot p){
	//No default file needed
	parent=p;
	setBorder(BorderFactory.createLineBorder(Color.black));
	width=500;
	height=350;
	ptsize=5;
	moving=-1;
	//	this.loadData(f);
	dim=3;
	ndata=0;
	this.setDefaultView();
	this.setColours();
	addMouseListener(new MouseAdapter() {
		public void mousePressed(MouseEvent e) {
		    mpress(e.getX(),e.getY());
		}
		public void mouseReleased(MouseEvent e) {
		    moving=-1;
		}
		public void mouseExit(MouseEvent e){
		    moving=-1;
		}
	    });

        addMouseMotionListener(new MouseAdapter() {
		public void mouseDragged(MouseEvent e) {
		    mdrag(e.getX(),e.getY());
		}
		public void mouseMoved(MouseEvent e) {
		    mmove(e.getX(),e.getY());
		}
	    });
    }


    public void ldaselect(int[] clst){
	int[] clno=new int[ndata];
	for(int i=0;i<ndata;i++){
	    clno[i]=clst[clss[i]];
	    //	    System.out.println(clno[i]+"");
	}
	this.lda(clno,4);
    }

    public void lda(int[] clusters,int numclust){
	double[][] icv=this.inclusterVariance(clusters,numclust);
	double[][] cmeans=new double[3][dim];
	double[] overallmean=new double[dim];
	for(int i=0;i<dim;i++){
	    icv[0][i]=1;
	    cmeans[0][i]=0;
	    cmeans[1][i]=0;
	    cmeans[2][i]=0;
	    overallmean[i]=0;
	}
	SquareMatrix M=new SquareMatrix(icv,dim);
	int[] cfr={0,0,0};
	for(int i=0;i<ndata;i++){
	    for(int j=0;j<dim;j++){
		overallmean[j]+=coeffs[i][j];
	    }
	    int clust=clusters[i];
	    if(clust<3){
		cfr[clust]++;
		for(int j=0;j<dim;j++){
		    cmeans[clust][j]+=coeffs[i][j];
		}
	    }
	}
	for(int i=0;i<dim;i++){
	    overallmean[i]/=ndata;
	    cmeans[0][i]/=cfr[0];
	    cmeans[1][i]/=cfr[1];
	    cmeans[1][i]-=overallmean[i];
	    cmeans[2][i]/=cfr[2];
	    cmeans[2][i]-=overallmean[i];
	    //	    System.out.println(cmeans[1][i]+","+cmeans[2][i]);
	}
	cmeans[1][0]=0;
	cmeans[2][0]=0;	    
	//	System.out.println("");
	double[] p1=M.solve(cmeans[1]);
	double[] p2=M.solve(cmeans[2]);
	double dot=0;
	double d1=0;
	double d2=0;
	for(int i=0;i<dim;i++){
	    dot+=p1[i]*p2[i];
	    d1+=p1[i]*p1[i];
	    d2+=p2[i]*p2[i];
	}
	double	subt=dot/(d1*d2);
	for(int i=0;i<dim;i++){
	    p2[i]-=p1[i]*subt;
	}
	double xmin=0;
	double ymin=0;
	double xmax=0;
	double ymax=0;
	for(int i=0;i<dim;i++){
	    if(p1[i]<xmin){
		xmin=p1[i];
	    }
	    if(p2[i]<ymin){
		ymin=p2[i];
	    }
	    if(p1[i]>xmax){
		xmax=p1[i];
	    }
	    if(p2[i]>ymax){
		ymax=p2[i];
	    }
	}

	for(int i=0;i<dim;i++){
	    points[i][0]=(p1[i]-xmin)*300/(xmax-xmin);
	    points[i][1]=(p2[i]-ymin)*300/(ymax-ymin);
	    //	    System.out.println(points[i][0]+","+points[i][1]);
	};
	repaint();
    }

    public double[][] inclusterVariance(int[] clusters,int numclust){
	int[] clfreq=new int[numclust];
	double[][][] sumsq=new double[numclust][dim][dim];
	double[][] means=new double[numclust][dim];
	for(int i=0;i<numclust;i++){
	    clfreq[i]=0;
	    for(int j=0;j<dim;j++){
		for(int k=0;k<dim;k++){
		    sumsq[i][j][k]=0;
		}
		means[i][j]=0;
	    }
	}
	for(int i=0;i<ndata;i++){	    
	    int clust=clusters[i];
	    clfreq[clust]++;
	    for(int j=0;j<dim;j++){
		for(int k=0;k<dim;k++){
		    sumsq[clust][j][k]+=coeffs[i][j]*coeffs[i][k];
		}
		means[clust][j]+=coeffs[i][j];
	    }
	}
	double[][]ans=new double[dim][dim];
	for(int j=0;j<dim;j++){
	    for(int k=0;k<dim;k++){
		ans[j][k]=0;
	    }
	}
	for(int i=0;i<numclust;i++){
	    for(int j=0;j<dim;j++){
		for(int k=0;k<dim;k++){
		    if(clfreq[i]>0){
			ans[j][k]+=(sumsq[i][j][k]-(means[i][j]*means[i][k])/clfreq[i]);
		    }
		}
	    }
	}
	return ans;
    }

    public void setMetaNum(int i){
	metanum=i;
    }

    public void setNumGreen(int i){
	numgreenpoints=i;
	repaint();
    }

    public void setDefaultView(){
	double real=Math.cos(2*Math.PI/dim);
	double imag=Math.sin(2*Math.PI/dim);
	double r=Math.sin(Math.PI/dim);
	double im=-Math.cos(Math.PI/dim);
	points=new double[dim][2];
	for(int i=0;i<dim;i++){
	    points[i][0]=(r+1.1)*100;
	    points[i][1]=(im+1.1)*100;
	    double rnew=r*real-im*imag;
	    im=r*imag+real*im;
	    r=rnew;
	    //	    System.out.println(points[i][0]+" "+points[i][1]+"i");
	};
    }

    public void setCTSColours(){
	//Sets the colours for metadata on a red-blue scale
	if(meta!=null){
	    double[] val=new double[ndata];
	    double sum=0;
	    double sumsq=0;
	    int nclss=0;
	    for(int i=0;i<ndata;i++){
		int k=meta.classes[metanum][i];
		val[k]=Double.parseDouble(meta.metaclasses[metanum][k]);
		if(k>nclss){
		    nclss=k;
		}
		sum+=val[k];
		sumsq+=val[k]*val[k];
	    }
	    double lower=sum/ndata-2*Math.sqrt(sumsq/ndata-sum*sum/ndata/ndata);
	    double diff=4*Math.sqrt(sumsq/ndata-sum*sum/ndata/ndata);
	    for(int j=0;j<nclss;j++){
		double rprop=(val[j]-lower)/diff;
		if(rprop<0){
		    rprop=0;
		}else if(rprop>1){
		    rprop=1;
		}
		double bprop=1-rprop;
		colours[j]=new Color((int)(rprop*255),20,(int)(bprop*255));
	    }
	}
    }

    public void setColours(){
	colours=new Color[dim+ndata];
	int pred=17;
	int pgreen=97;
	int pblue=197;
	for(int i=0;i<dim+ndata;i++){
	    colours[i]=new Color((pred*(i+1))%256,(pgreen*(i+1))%256,(pblue*(i+1))%256);
	}
    }

    public void loadData(File f){
	Scanner stream=null;
	try{
	    stream=new Scanner(new BufferedReader(new FileReader(f)));
	} catch(FileNotFoundException e) {
	    System.err.println("File: "+f.getName()+" not found!");
	} 
	try{
	    int tmpdim=stream.nextInt();
	    int tmpndata=stream.nextInt();
	    //	    System.out.println("dimension "+tmpdim+" data "+tmpndata);
	    double[][] tmpCoeffs=new double[tmpndata][tmpdim];
	    int[] tmpclss=new int[tmpndata];
	    for(int i=0;i<tmpndata;i++){
		tmpclss[i]=0;//stream.nextInt();
		for(int j=0;j<tmpdim;j++){
		    tmpCoeffs[i][j]=stream.nextDouble();
		}
	    }
	    stream.close();
	    clss=tmpclss;
	    coeffs=tmpCoeffs;
	    if(dim!=tmpdim){
		dim=tmpdim;
		points=new double[dim][2];
		this.setDefaultView();
	    };
	    dim=tmpdim;
	    ndata=tmpndata;
	    this.setColours();
	    repaint();
	}finally{
	}
	//catch(IOException e){
	 //   System.out.println("Failed to read data from file.");
	 //   try{
	//	stream.close();
	//   }catch(IOException ex){
	//	System.out.println("Can't close file!!!!!!!.");
	//    }       
	//}
    }

    public void loadView(File f){
	Scanner stream=null;
	try{
	    stream=new Scanner(new BufferedReader(new FileReader(f)));
	} catch(FileNotFoundException e) {
	    System.err.println("File: "+f.getName()+" not found!");
	} 
	try{
	    int tmpdim=stream.nextInt();
	    if(tmpdim!=dim){
		System.out.println("Dimension does not match data dimension.");
		stream.close();
		return;
	    }
	    double[][] tmpPoints=new double[dim][2];
	    for(int i=0;i<dim;i++){
		tmpPoints[i][0]=stream.nextDouble();
		tmpPoints[i][1]=stream.nextDouble();
	    }
	    stream.close();
	    points=tmpPoints;
	    repaint();
	    /*	}catch(IOException e){
	    System.out.println("Failed to read data from file.");
	    stream.close();
	    }*/
	}finally{
	}
    }

    public void saveView(File f){
	PrintWriter stream=null;
	try{
	    stream=new PrintWriter(new BufferedWriter(new FileWriter(f)));
	    stream.println(dim);
	    for(int i=0;i<dim;i++){
		stream.print(points[i][0]);
		stream.print(" ");
		stream.println(points[i][1]);
	    }
	}catch(IOException e){
	    System.err.println("Error writing to file: "+f.getName());
	    System.err.println(e.getMessage());
	}finally{
	    if(stream!=null){
		stream.close();
	    }
	}
    }




    private void mpress(int x,int y){
	//check all points to see if one is clicked on.
	for(int i=0;i<dim;i++){
	    if(((x-points[i][0])*(x-points[i][0])<=4&&(y-points[i][1])*(y-points[i][1])<ptsize*ptsize)||((y-points[i][1])*(y-points[i][1])<=4&&(x-points[i][0])*(x-points[i][0])<=ptsize*ptsize)){
		moving=i;
	    }
	}
    }

    private void mdrag(int x,int y){
	if(moving>=0){
	    if(x>0&&y>0&&x<this.getWidth()&&y<this.getHeight()){
		points[moving][0]=x;
		points[moving][1]=y;
		repaint();
	    }else{
		moving=-1;
	    };
	};
    }

    private void mmove(int x,int y){
	if(moving==-1){
	    boolean onpoint=false;
	    for(int i=0;i<ndata;i++){
		double px=0;
		double py=0;
		for(int j=0;j<dim;j++){
		    px+=coeffs[i][j]*points[j][0];
		    py+=coeffs[i][j]*points[j][1];
		}
		if(((x-px)*(x-px)+(y-py)*(y-py))<ptsize*ptsize){
		    onpoint=true;
		    String clname;
		    if(meta==null){
			clname=String.valueOf(clss[i]);
		    }else{
			clname=meta.metaclasses[metanum][meta.classes[metanum][i]];
		    }
		    String dname;
		    if(meta==null){
			dname=String.valueOf(i);
		    }else{
			dname=meta.names[i];
		    }
		    int msglen=10;
		    if(dname.length()<10){
			msglen=dname.length();
		    }
		    parent.message.setText("point: "+dname.substring(0,msglen));
		    msglen=10;
		    if(clname.length()<10){
			msglen=clname.length();
		    }
		    parent.message2.setText("class: "+clname.substring(0,msglen));
		    if(meta!=null){
			for(int j=0;j<meta.nummeta;j++){
			    String mnm=meta.metaclasses[j][meta.classes[j][i]];
			    msglen=10;
			    if(mnm.length()<10){
				msglen=mnm.length();
			    }
			    if(j<metanum){
				parent.extmessages[j].setText((j+1)+"   : "+mnm.substring(0,msglen));
			    }
			    if(j>metanum){
				parent.extmessages[j-1].setText((j+1)+"   : "+mnm.substring(0,msglen));
			    }
			}
		    }
		}
	    }
	    if(!onpoint){
		parent.message.setText("              ");
		parent.message2.setText("                     ");
		if(meta!=null&&parent.extmessages!=null){
		    for(int j=0;j<meta.nummeta-1;j++){
			parent.extmessages[j].setText("                     ");
		    };
		};
	    }
	}
    }

    public Dimension getPreferredSize(){
	return new Dimension(width,height);
    }

    public void chooseMeta(int m){
	if(m>=0&m<meta.nummeta){
	    metanum=m;
	    clss=meta.classes[m];
	    repaint();
	}
    }

    public void paintComponent(Graphics g){
	super.paintComponent(g);
	Graphics2D g2=(Graphics2D)g;
	for(int i=0;i<ndata;i++){
	    double x=0;
	    double y=0;
	    for(int j=0;j<dim;j++){
		x+=coeffs[i][j]*points[j][0];
		y+=coeffs[i][j]*points[j][1];
	    }
	    g2.setPaint(colours[clss[i]]);
	    g2.fill(new Ellipse2D.Double(x-ptsize,y-ptsize,2*ptsize,2*ptsize));
	}
	for(int i=0;i<dim;i++){
	    if(i<numgreenpoints){
		g2.setPaint(Color.green);
	    }else{
		g2.setPaint(Color.black);
	    }
	    g2.fill(new Rectangle2D.Double(points[i][0]-ptsize,points[i][1]-1,2*ptsize,2));
	    g2.fill(new Rectangle2D.Double(points[i][0]-1,points[i][1]-ptsize,2,2*ptsize));
	}
	if(meta!=null){
	    for(int m=0;m<meta.numlinks;m++){
		int i=meta.links[2*m];
		int j=meta.links[2*m+1];
		
		double x1=0;
		double y1=0;
		for(int k=0;k<dim;k++){
		    x1+=coeffs[i][k]*points[k][0];
		    y1+=coeffs[i][k]*points[k][1];
		}
		double x2=0;
		double y2=0;
		for(int k=0;k<dim;k++){
		    x2+=coeffs[j][k]*points[k][0];
		    y2+=coeffs[j][k]*points[k][1];
		}
		g2.setPaint(meta.linkcolours[meta.linktype[m]]);
		g2.draw(new Line2D.Double(x1,y1,x2,y2));
	    }	    
	}
    }

    public void exportTops(File f){
	double maxx=0;
	double maxy=0;
	double minx=0;
	double miny=0;
	for(int i=0;i<dim;i++){
	    if(points[i][0]>maxx){
		maxx=points[i][0];
	    };
	    if(points[i][1]>maxy){
		maxy=points[i][1];
	    };
	    if(points[i][0]<minx){
		minx=points[i][0];
	    };
	    if(points[i][1]<miny){
		miny=points[i][1];
	    };
	};
	maxx+=5;
	maxy+=5;
	minx-=5;
	miny-=5;

	int numlegend=ndata;
	if(colours.length<numlegend){
	    numlegend=colours.length;
	}
	int lwchar=0;
	for(int cl=0;cl<ndata&&cl<colours.length;cl++){
	    if(meta.metaclasses[metanum][cl]==null){
		numlegend--;
	    }else if(meta.metaclasses[metanum][cl].length()>lwchar){
		lwchar=meta.metaclasses[metanum][cl].length();
	    }
	}
	int legendwidth=(60+16*lwchar);
	int legendheight=(30*numlegend+26);
	int offset=legendheight+10;
	double rescalex=1000/(maxx-minx);
	double rescaley=(1000-offset)/(maxy-miny);
	try{
	    //	    File template=new File("template.eps");
	    BufferedWriter stream=new BufferedWriter(new FileWriter(f));
	    BufferedReader templ=new BufferedReader(new InputStreamReader(getClass().getResourceAsStream("template.eps")));
	    String line;
	    while((line =templ.readLine())!=null){
		stream.write(line);
		stream.newLine();
	    }
	    templ.close();
	    stream.write("gsave");
	    stream.newLine();
	    stream.write("0 0 0 setrgbcolor");
	    stream.newLine();
	    for(int i=0;i<dim;i++){
		stream.write((points[i][0]-minx)*rescalex+" "+((maxy-points[i][1])*rescaley+offset)+" plotvt");
		stream.newLine();
	    };
	    for(int cl=0;cl<ndata&&cl<numlegend;cl++){
		stream.write(((double) colours[cl].getRed())/255+" "+((double) colours[cl].getGreen())/255+" "+((double) colours[cl].getBlue())/255+" setrgbcolor");
		stream.newLine();
		for(int i=0;i<ndata;i++){
		    if(clss[i]==cl){
			double x=0;
			double y=0;
			for(int j=0;j<dim;j++){
			    x+=coeffs[i][j]*points[j][0];
			    y+=coeffs[i][j]*points[j][1];
			}
			stream.write((x-minx)*rescalex+" "+((maxy-y)*rescaley+offset)+" plotpt");
			stream.newLine();
		    }
		}
		for(int m=0;m<meta.numlinks;m++){
		    int i=meta.links[2*m];
		    int j=meta.links[2*m+1];
		    
		    double x1=0;
		    double y1=0;
		    for(int k=0;k<dim;k++){
			x1+=coeffs[i][k]*points[k][0];
			y1+=coeffs[i][k]*points[k][1];
		    }
		    double x2=0;
		    double y2=0;
		    for(int k=0;k<dim;k++){
			x2+=coeffs[j][k]*points[k][0];
			y2+=coeffs[j][k]*points[k][1];
		    }
		stream.write(((double) meta.linkcolours[meta.linktype[m]].getRed())/255+" "+((double)  meta.linkcolours[meta.linktype[m]].getGreen())/255+" "+((double) meta.linkcolours[meta.linktype[m]].getBlue())/255+" setrgbcolor");
		stream.newLine();
		stream.write((x1-minx)*rescalex+" "+((maxy-y1)*rescaley+offset)+" moveto");
		stream.newLine();
		stream.write((x2-minx)*rescalex+" "+((maxy-y2)*rescaley+offset)+" lineto stroke");
		stream.newLine();

		}
	    }

	    stream.write("grestore");
	    stream.newLine();
	    stream.write("/legendwidth "+legendwidth+" def");
	    stream.newLine();
	    stream.write("/legendheight "+legendheight+" def");
	    stream.newLine();
	    stream.write((1000-legendwidth)/2+" 5 legendbox");
	    stream.newLine();
	    for(int cl=0;cl<ndata&&cl<colours.length&&meta.metaclasses[metanum][cl]!=null;cl++){
		stream.write("("+meta.metaclasses[metanum][cl]+") "+(((double)colours[cl].getRed())/255)+" "+(((double)colours[cl].getGreen())/255)+" "+(((double)colours[cl].getBlue())/255)+" legenditem");
		stream.newLine();
	    }
	    stream.close();
	}catch(FileNotFoundException e){
	    System.err.println("File: \"template.eps\" not found!");
	}catch(IOException e){
	}
    }
}

class LDAchooser extends AbstractAction{
    JFrame window;
    JButton OK;
    JButton cancel;
    JRadioButton[][] buttons;
    ButtonGroup[] buttongroups;
    Splot plot;
    int[] clst;    //Chooses the classes to discriminate between
    int numclass=0;
    public LDAchooser(Splot p){
	plot=p;
	window=new JFrame("Choose classes");
	JPanel panel=new JPanel(new GridBagLayout());	
 	GridBagConstraints c = new GridBagConstraints();
	numclass=0;
	for(int i=0;i<plot.ndata;i++){
	    if(plot.clss[i]+1>numclass){
		numclass=plot.clss[i]+1;
	    }
	}
	buttons=new JRadioButton[3][numclass];
	buttongroups=new ButtonGroup[numclass];
	clst=new int[plot.ndata];
	for(int i=0;i<plot.ndata;i++){
	    clst[i]=3;
	}
 	c.fill = GridBagConstraints.HORIZONTAL;
	c.gridx=1;
	c.gridy=0;
	panel.add(new JLabel("class1"),c);
	c.gridx=2;
	c.gridy=0;
	panel.add(new JLabel("class2"),c);
	c.gridx=3;
	c.gridy=0;
	panel.add(new JLabel("class3"),c);
	for(int i=0;i<numclass;i++){
	    buttons[0][i]=new JRadioButton();
	    buttons[1][i]=new JRadioButton();
	    buttons[2][i]=new JRadioButton();
	    buttongroups[i]=new ButtonGroup();
	    buttongroups[i].add(buttons[0][i]);
	    buttongroups[i].add(buttons[1][i]);
	    buttongroups[i].add(buttons[2][i]);
	    buttons[0][i].addActionListener(this);
	    buttons[1][i].addActionListener(this);
	    buttons[2][i].addActionListener(this);
	    c.gridy=i+1;
	    c.gridx=0;
	    panel.add(new JLabel(plot.meta.metaclasses[plot.metanum][i]),c);
	    c.gridx=1;
	    panel.add(buttons[0][i],c);
	    c.gridx=2;
	    panel.add(buttons[1][i],c);
	    c.gridx=3;
	    panel.add(buttons[2][i],c);
	}
	c.gridy=numclass+1;
	OK=new JButton("OK");
	OK.addActionListener(this);
	panel.add(OK,c);
	cancel=new JButton("Cancel");
	cancel.addActionListener(this);
	c.gridx=0;
	panel.add(OK,c);
	c.gridx=2;
	panel.add(cancel,c);
	window.getContentPane().add(panel,BorderLayout.CENTER);
	window.pack();
	window.setVisible(true);
    }

    public void actionPerformed(ActionEvent e){
	if(e.getSource()==OK){	    
	    plot.ldaselect(clst);
	    window.setVisible(false);
	}else if(e.getSource()==cancel){
	    //window.unpack();
	    //	    System.out.printl
	    window.setVisible(false);
	}else{
	    for(int i=0;i<numclass;i++){
		if(e.getSource()==buttons[0][i]){
		    if(clst[i]==0){
			clst[i]=3;
			buttons[0][i]=new JRadioButton();
			buttongroups[i].add(buttons[0][i]);
		    }else{
			clst[i]=0;
		    }
		}else if(e.getSource()==buttons[1][i]){
		    if(clst[i]==1){
			clst[i]=3;
			buttons[1][i]=new JRadioButton();
			buttongroups[i].add(buttons[1][i]);
		    }else{
			clst[i]=1;
		    }
		}else if(e.getSource()==buttons[2][i]){
		    if(clst[i]==0){
			clst[i]=3;
			buttons[0][i]=new JRadioButton();
			buttongroups[i].add(buttons[2][i]);
		    }else{
			clst[i]=2;
		    }
		}
	    }
	}
    }    


}


class LDAAction extends AbstractAction{
    SimplePlot top;
    public LDAAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	int[] clusters=new int[top.plot.ndata];
	int ncl=0;
	for(int i=0;i<top.plot.ndata;i++){
	    clusters[i]=top.plot.meta.classes[top.plot.metanum][i];
	    if(clusters[i]>ncl){
		ncl=clusters[i]+1;
	    }
	}
	//	top.plot.lda(clusters,ncl);
	LDAchooser choose=new LDAchooser(top.plot);
    }


}

class LoadDataAction extends AbstractAction{
    SimplePlot top;
    JFileChooser fc;
    public LoadDataAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	fc=new JFileChooser(".");
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	fc.setApproveButtonText("Load");
	fc.setDialogTitle("Load Data");
	int decision=fc.showOpenDialog(top.mainWindow);
	if(decision==JFileChooser.APPROVE_OPTION){
	    File file=fc.getSelectedFile();
	    if(file.isDirectory()){
		//		fc.setDirectory(file);
	    }else{
		top.plot.loadData(file);
	    }
	}
    }
}

class LoadMetaDataAction extends AbstractAction{
    SimplePlot top;
    JFileChooser fc;
    public LoadMetaDataAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	fc=new JFileChooser(".");
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	fc.setApproveButtonText("Load");
	fc.setDialogTitle("Load MetaData");
	int decision=fc.showOpenDialog(top.mainWindow);
	if(decision==JFileChooser.APPROVE_OPTION){
	    File file=fc.getSelectedFile();
	    if(file.isDirectory()){
		//		fc.setDirectory(file);
	    }else{
		top.plot.meta=new metadata(file,top.plot.ndata);
		top.setMetaDataNumber(top.plot.meta.nummeta);
		top.plot.chooseMeta(0);
		top.setLegend();
	    }
	}
    }
}

class SetCTSColourAction extends AbstractAction{
    SimplePlot top;
    public SetCTSColourAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
    }
    public void actionPerformed(ActionEvent e){
	top.plot.setCTSColours();
    }
}



class LoadViewAction extends AbstractAction{
    SimplePlot top;
    JFileChooser fc;
    public LoadViewAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	fc=new JFileChooser(".");
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	fc.setApproveButtonText("Load");
	fc.setDialogTitle("Load View");
	int decision=fc.showOpenDialog(top.mainWindow);
	if(decision==JFileChooser.APPROVE_OPTION){
	    File file=fc.getSelectedFile();
	    if(file.isDirectory()){
		//		fc.setDirectory(file);
	    }else{
		top.plot.loadView(file);
	    }
	}
    }
}


class SaveViewAction extends AbstractAction{
    SimplePlot top;
    JFileChooser fc;
    public SaveViewAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	fc=new JFileChooser(".");
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	int decision=fc.showSaveDialog(top.mainWindow);
	if(decision==JFileChooser.APPROVE_OPTION){
	    File file=fc.getSelectedFile();
	    if(file.isDirectory()){
		//		fc.setDirectory(file);
	    }else{
		top.plot.saveView(file);
	    }
	}
    }
}

class ExportAction extends AbstractAction{
    SimplePlot top;
    JFileChooser fc;
    public ExportAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	fc=new JFileChooser(".");
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	int decision=fc.showSaveDialog(top.mainWindow);
	if(decision==JFileChooser.APPROVE_OPTION){
	    File file=fc.getSelectedFile();
	    if(file.isDirectory()){
		//		fc.setDirectory(file);
	    }else{
		top.plot.exportTops(file);
	    }
	}
    }
}


class SetMetaDataAction extends AbstractAction{
    SimplePlot top;
    JFileChooser fc;
    public SetMetaDataAction(String text, String desc, int mnemonic,SimplePlot b){
	top=b;
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	fc=new JFileChooser(".");
	//fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
	//	fc.setFileFilter(new GameFilter());
    }
    public void actionPerformed(ActionEvent e){
	fc.setApproveButtonText("Load");
	fc.setDialogTitle("Load MetaData");
	int decision=fc.showOpenDialog(top.mainWindow);
	if(decision==JFileChooser.APPROVE_OPTION){
	    File file=fc.getSelectedFile();
	    if(file.isDirectory()){
		//		fc.setDirectory(file);
	    }else{
		top.plot.meta=new metadata(file,top.plot.ndata);
		top.setLegend();
	    }
	}
    }
}



class HelpAction extends AbstractAction{
    boolean loaded;
    JScrollPane display;
    public HelpAction(String text, String desc,int mnemonic){
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	loaded=false;    
    }
    public void actionPerformed(ActionEvent e){
	JFrame helpWindow=new JFrame("help");
	//	if(!loaded){
	    JEditorPane text=new JEditorPane();
	    java.net.URL helpText=HelpAction.class.getResource("help.html");
	    if(helpText==null){
		System.err.println("Sorry: no help available.");
	    }else{
		try{
		    text.setPage(helpText);
		}catch(IOException f){
		    System.err.println("Couldn't open help file.");
		    return;
		}
		display=new JScrollPane(text);
		display.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		display.setPreferredSize(new Dimension(500,300));
		display.setMinimumSize(new Dimension(20,20));
	    }
	    //	}
	    text.setEditable(false);
	    JPanel pane=new JPanel(new GridLayout(1,0));
	    pane.add(display,BorderLayout.CENTER);
	helpWindow.getContentPane().add(pane,BorderLayout.CENTER);
	helpWindow.pack();
	helpWindow.setVisible(true);
    }
}

class ChooseMetaAction extends AbstractAction{
    int val;
    SimplePlot top;
    public ChooseMetaAction(String text, String desc,int mnemonic,int v,SimplePlot t){
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	top=t;
	val=v;
    }
    public void actionPerformed(ActionEvent e){
	top.plot.chooseMeta(val);
	top.setLegend();
    }
}

class setNumGreenPointsAction extends AbstractAction{
    SimplePlot top;
    public setNumGreenPointsAction(String text, String desc,int mnemonic,SimplePlot t){
	putValue(NAME,text);
	putValue(SHORT_DESCRIPTION,desc);
	top=t;
    }
    public void actionPerformed(ActionEvent e){
	int val=0;
	String v=(String)JOptionPane.showInputDialog(top.plot,"Number of Points to highlight:","Highlight Points",JOptionPane.PLAIN_MESSAGE,null,null,null);
	val=Integer.parseInt(v);
	top.plot.setNumGreen(val);
    }
}


class SetColourAction extends MouseAdapter{
    int cnum;
    SimplePlot top;
    //    JColorChooser chooser;
    public SetColourAction(int cn,SimplePlot t,Color current){
	//	putValue(NAME,text);
	//	putValue(SHORT_DESCRIPTION,desc);
	super();
	top=t;
	cnum=cn;
	//	chooser=new JColorChooser(current);
    }
    public void mouseClicked(MouseEvent e){
	Color ncl=JColorChooser.showDialog(top.mainWindow,"Choose colour",top.plot.colours[cnum]);//chooser.getColor();
	top.plot.colours[cnum]=ncl;
	top.plot.repaint();
	top.setLegend();
    }
}



public class SimplePlot{

    //    File file;
    JFrame mainWindow;
    Splot plot;
    JPanel legend;
    JLabel message;
    JLabel message2;
    JPanel extmboard;
    JLabel extmessages[];
    //    WindowCloser endProgram;
    int mtnum;
    ButtonGroup bg;
    JMenu metamenu;
    static File commandLineData;
    static File commandLineMetadata;
    static File directory;
    JPanel superpanel;

    public SimplePlot(){
	//	endProgram=new WindowCloser();
	//	loadSettings();
	mtnum=0;
	//	File datafile=new File("DefaultDataFile");
	plot=new Splot(this);

	message=new JLabel("                                ");
	message.setPreferredSize(new Dimension(135,17));
	message2=new JLabel("                                ");
	message2.setPreferredSize(new Dimension(135,17));
	superpanel=new JPanel(new GridBagLayout());	
 	GridBagConstraints c = new GridBagConstraints();
 	c.fill = GridBagConstraints.HORIZONTAL;

 	c.gridwidth=5;
 	c.gridheight=5;
	c.gridx = 0;
 	c.gridy = 0;
	c.fill=GridBagConstraints.BOTH;
	c.weightx=1;
	c.weighty=1;
 	superpanel.add(plot,c);
	c.fill=GridBagConstraints.NONE;
	c.weightx=0;
	c.weighty=0;
	c.gridx=5;
	c.gridheight=1;
	superpanel.add(message,c);
	c.gridy=2;
	c.gridheight=1;
	superpanel.add(message2,c);
	c.gridy=3;
	c.gridheight=1;
	extmboard=new JPanel(new GridBagLayout());
	superpanel.add(extmboard,c);
	c.gridy=4;
	c.gridheight=1;
	c.weighty=100;
	c.fill=GridBagConstraints.BOTH;
	legend=new JPanel(new GridBagLayout());	
	//	legend.setPreferredSize(new Dimension(135,300));
	legend.setAlignmentX(Component.LEFT_ALIGNMENT);
	legend.setBorder(BorderFactory.createLineBorder(Color.black));
	superpanel.add(legend,c);
	mainWindow=new JFrame("Simple Plot");
	this.setUpMenus();
	//	mainWindow.addWindowListener(endProgram);
	mainWindow.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	mainWindow.getContentPane().add(superpanel,BorderLayout.CENTER);
	mainWindow.pack();
	mainWindow.setVisible(true);
	if(commandLineData!=null){
	    plot.loadData(commandLineData);   
	}
	if(commandLineMetadata!=null){
	    plot.meta=new metadata(commandLineMetadata,plot.ndata);
	    this.setMetaDataNumber(plot.meta.nummeta);
	    plot.chooseMeta(0);
	    this.setLegend();
	}
    }

    public void setLegend(){
	legend.removeAll();
	Color[] colours=plot.colours;
	String[] labels=plot.meta.metaclasses[plot.metanum];
	GridBagConstraints c=new GridBagConstraints();
 	c.fill = GridBagConstraints.HORIZONTAL;

 	c.gridwidth=1;
 	c.gridheight=1;
	c.gridx=0;
	c.gridy=0;
	c.weightx=0;
	c.weighty=0;
	for(int i=0;i<labels.length&&labels[i]!=null;i++){
	    c.gridy=i;
	    JPanel sample=new JPanel();
	    sample.setBackground(colours[i]);
	    sample.setPreferredSize(new Dimension(13,13));
	    sample.addMouseListener(new SetColourAction(i,this,colours[i]));
	    legend.add(sample,c);
	}
	c.gridx=1;
	c.gridwidth=3;
	c.weightx=1;
	for(int i=0;i<labels.length&&labels[i]!=null;i++){
	    c.gridy=i;
	    int msglen=15;
	    if(labels[i].length()<msglen){
		msglen=labels[i].length();
	    }
	    JLabel text=new JLabel(labels[i].substring(0,msglen));
	    legend.add(text,c);
	    c.gridy=i+1;
	}
	c.weighty=1;
	legend.add(new JLabel(""),c);
	legend.repaint();
	mainWindow.validate();
	mainWindow.repaint();
    }

    public void setMetaDataNumber(int n){
	//	System.out.println("Setting number of metadata to "+n);
	if(bg==null){
	    bg=new ButtonGroup();
	}
	for(int i=mtnum;i<n;i++){
	    //	    System.out.println("creating button "+i);
	    JRadioButtonMenuItem mIt=new JRadioButtonMenuItem(""+(i+1),false);
	    mIt.setAction(new ChooseMetaAction(""+(i+1),"",0,i,this));
	    metamenu.add(mIt);
	    bg.add(mIt);		    
	}
	extmessages=new JLabel[n-1];
	GridBagConstraints c = new GridBagConstraints();
	c.fill = GridBagConstraints.HORIZONTAL;
	
	c.gridwidth=1;
	c.gridheight=1;	    
	c.fill=GridBagConstraints.BOTH;
	c.weightx=1;
	c.weighty=1;
	for(int i=0;i<n-1;i++){
	    extmessages[i]=new JLabel("                 ");
	    extmessages[i].setPreferredSize(new Dimension(135,17));      
	    c.gridx = 0;
	    c.gridy = i;
	    extmboard.add(extmessages[i],c);
	}
	mtnum=n;
    }
    public void setUpMenus(){
	JMenuBar menuBar=new JMenuBar();
	JMenu menu=new JMenu("File");
	menu.setMnemonic(KeyEvent.VK_F);
	menu.getAccessibleContext().setAccessibleDescription("File options");
	menuBar.add(menu);
	JMenuItem menuItem=new JMenuItem("Load Data",KeyEvent.VK_D);
	
	menuItem.setAction(new LoadDataAction("Load Data","Loads data from a file.",0,this));
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK));
	menu.add(menuItem);

	menuItem = new JMenuItem("Load View",KeyEvent.VK_V);
	menuItem.setAction(new LoadViewAction("Load View","Loads view from previously saved file",0,this));
	menu.add(menuItem);
	
	menuItem=new JMenuItem("Save",KeyEvent.VK_S);
	menuItem.setAction(new SaveViewAction("Save","Saves the current view",0,this));
	menu.add(menuItem);
				
	menuItem=new JMenuItem("Export",KeyEvent.VK_E);
	menuItem.setAction(new ExportAction("Export","Exports the current view to postscript file",0,this));
	menu.add(menuItem);
				
	menuItem=new JMenuItem("Highlight Points",KeyEvent.VK_E);
	menuItem.setAction(new setNumGreenPointsAction("Highlight Points","highlights a number of dimensions",0,this));
	menu.add(menuItem);

	menuItem = new JMenuItem("Load View",KeyEvent.VK_V);
	menuItem.setAction(new SetCTSColourAction("Continuous Colour","Uses a colour scale to display continous metadata",0,this));
	menu.add(menuItem);
					

	metamenu=new JMenu("MetaData");
	menuBar.add(metamenu);	      

	menuItem=new JMenuItem("Load",KeyEvent.VK_M);
	menuItem.setAction(new LoadMetaDataAction("Load","LoadsMetaData",0,this));
	metamenu.add(menuItem);



//	menu=new JMenu("LDA");
//	menuBar.add(menu);
//	menuItem=new JMenuItem("LDA",KeyEvent.VK_H);
//	menuItem.setAction(new LDAAction("LDA","Computes a good view using linear discriminant analysis",0,this));
//	menu.add(menuItem);


	menu=new JMenu("Help");
	menuBar.add(menu);
	menuItem=new JMenuItem("Show help",KeyEvent.VK_H);
	menuItem.setAction(new HelpAction("Help","Displays help information",0));
	menu.add(menuItem);
	//	menu=new JMenu("Options");
	//	menuBar.add(menu);
	mainWindow.setJMenuBar(menuBar);
    }

    public static void createGUI(){
	JFrame.setDefaultLookAndFeelDecorated(true);
	//	game.runner=new RunGame(game.players[0],game.players[1],game);
	//	game.mainThread=new Thread(game.runner);
	//	game.mainThread.start();
    }

    public static void main(String[] args){
	if(args.length>2){
	    directory=new File(args[2]);
	}
	if(args.length>1){
	    commandLineMetadata=new File(args[1]);
	}
	if(args.length>0){
	    commandLineData=new File(args[0]);
	}else{
	    commandLineData=null;
	}
	javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
		SimplePlot plotwindow=new SimplePlot();
                plotwindow.createGUI();
            }
        });
	return;      
    }
}

